import os
import numpy as np
from PIL import Image

import torch
import onnxruntime as ort
import onnx
import ktc
import kneronnxopt

def softmax(x, axis=1):
    e_x = np.exp(x)
    return e_x / e_x.sum(axis=axis, keepdims=True)

def preprocess(input_file):
    image = Image.open(input_file).convert("RGB")
    image = np.array(image.resize((224,224), Image.BILINEAR)) / 255
    image = np.transpose(image, (2, 0, 1))[None]
    return image

def inference_comparison(a_probs, b_probs, a_labels, b_labels):
    prob_diff = np.abs(a_probs - b_probs).max()
    label_diff = (a_labels != b_labels).sum() / len(a_labels)
    return prob_diff, label_diff

def save_results(labels, probs, model_name, file_name, mode='a'):
    print(f"{model_name} Labels: {labels} \n")
    print(f"{model_name} Probabilities: {probs} \n\n")
    with open(file_name, mode) as f:
        f.write(f"{model_name} Labels: {labels} \n")
        f.write(f"{model_name} Probabilities: {probs} \n\n")
    

def save_differences(labels, probs, model_name, file_name, mode='a'):
    print(f"Label difference ratio for {model_name}: {labels} \n")
    print(f"Probability difference for {model_name}: {probs} \n\n")
    with open(file_name, mode) as f:
        f.write(f"Label difference ratio for {model_name}: {labels} \n")
        f.write(f"Probability difference for {model_name}: {probs} \n\n")

results_file = "results.txt"
raw_images = os.listdir("./images/")
input_images = [preprocess("./images/" + image_name) for image_name in raw_images]

# 1. Use ONNX Runtime to infer original onnx
onnx_model = "mobilenet_v2-b0353104.onnx"
onnx_session = ort.InferenceSession(onnx_model)
input_name = onnx_session.get_inputs()[0].name
onnx_predictions = []
for input_data in input_images:
    input_data = np.expand_dims(input_data, axis=0).astype(np.float32) 
    input_data = input_data.squeeze(axis=0) 
    pred = onnx_session.run(None, {input_name: input_data})[0]
    onnx_predictions.append(pred)
    
onnx_from_torch_predictions = np.vstack(onnx_predictions).reshape(len(input_images),-1)
onnx_predictions = softmax(onnx_from_torch_predictions)
onnx_labels = onnx_predictions.argmax(axis=1)
onnx_probs = onnx_predictions.max(axis=1)
save_results(onnx_labels, onnx_probs, "original onnx", file_name=results_file, mode='w')

# 2. optimized onnx model
torch_version = torch.__version__.split(".")[:2]
if torch_version == ['2', '3']:
    optonnx = kneronnxopt.optimize(onnx.load(onnx_model))
elif torch_version == ['1', '7']:
    result_m = ktc.onnx_optimizer.torch_exported_onnx_flow(onnx.load(onnx_model))
    optonnx = ktc.onnx_optimizer.onnx2onnx_flow(result_m, eliminate_tail=True, opt_matmul=False)
optonnx_model = "mobilenet_v2-b0353104_opt.onnx"
onnx.save(optonnx, optonnx_model)

# 2.1. optimized onnx model using kneron inference
optonnx_predictions = []
for input_data in input_images:
    pred = ktc.kneron_inference(input_data, onnx_file=optonnx_model, input_names=["input.1"])   
    optonnx_predictions += pred
optonnx_predictions = np.vstack(optonnx_predictions).reshape(len(input_images),-1)
optonnx_predictions = softmax(optonnx_predictions)
optonnx_labels = optonnx_predictions.argmax(axis=1)
optonnx_probs = optonnx_predictions.max(axis=1)
save_results(optonnx_labels, optonnx_probs, "optimized onnx (kneron inference)", file_name=results_file)

# 2.2. optimized onnx model using ONNX Runtime
onnx_session = ort.InferenceSession(optonnx_model)
input_name = onnx_session.get_inputs()[0].name
optonnx_predictions_onnxinf = []
for input_data in input_images:
    input_data = np.expand_dims(input_data, axis=0).astype(np.float32)
    input_data = input_data.squeeze(axis=0) 
    pred = onnx_session.run(None, {input_name: input_data})[0]
    optonnx_predictions_onnxinf.append(pred)
optonnx_predictions_onnxinf = np.vstack(optonnx_predictions_onnxinf).reshape(len(input_images),-1)
optonnx_predictions_onnxinf = softmax(optonnx_predictions_onnxinf)
optonnx_labels_onnxinf = optonnx_predictions_onnxinf.argmax(axis=1)
optonnx_probs_onnxinf = optonnx_predictions_onnxinf.max(axis=1)
save_results(optonnx_labels_onnxinf, optonnx_probs_onnxinf, "optimized onnx (ort inference)", file_name=results_file)

######### Compare.1 original onnx vs. optimized onnx (kneron inference) #########
prob23_diff, label23_diff = inference_comparison(onnx_predictions, optonnx_predictions, onnx_labels, optonnx_labels)
save_differences(label23_diff, prob23_diff, "original onnx vs. optimized onnx (kneron inference)", file_name=results_file)
# Note: The difference of the two models are huge here!!!

######### Compare.2 original onnx vs. optimized onnx (ort inference) #########
prob24_diff, label24_diff = inference_comparison(onnx_predictions, optonnx_predictions_onnxinf, onnx_labels, optonnx_labels_onnxinf)
save_differences(label24_diff, prob24_diff, "original onnx vs. optimized onnx (ort inference)", file_name=results_file)
# Note: The difference of the two models is good.

######### Compare.3 optimized onnx: kneron inference vs. ort inference #########
prob34_diff, label34_diff = inference_comparison(optonnx_predictions, optonnx_predictions_onnxinf, optonnx_labels, optonnx_labels_onnxinf)
save_differences(label34_diff, prob34_diff, "optimized onnx: kneron inference vs. ort inference", file_name=results_file)
# Note: The difference of the inference functions from the same model are huge here!!!